# This code uses the __getattr__ magic method to dynamically import cli and stdin only
# when they are accessed. Importing them directly at the top of __init__.py interfere with
# the dependencies installation when huckle is installed via pip.
#
# If developing using pip editable installs on huckle and/or code referencing huckle
# as a library, an accidental working directory subfolder named huckle, anywhere you happen
# to be when executing related code, may prevent huckle from being imported properly as a library.
# This is a known side effect of PEP660 and import overshadowing on pip editable installs.
# This can be mitigated by ensuring no such subfolder exists where you are working, and may be mitigated
# by changing the editable installs to compability mode.
#
# (e.g. pip install -e . --config-settings editable_mode=compat).
#
# See https://github.com/pypa/setuptools/issues/3548.

__all__ = ['cli', 'stdin']

def __getattr__(name):
    if name in __all__:
        from .huckle import cli, stdin
        return locals()[name]
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import subprocess
import os
import pytest

def test_hfm_cp(cleanup):
    setup = """
    #!/bin/bash
    set -x
    export NOTIFY_SOCKET=

    export HCLI_CORE_HOME=~/.hcli_core_test
    export HUCKLE_HOME=~/.huckle_test
    eval $(huckle env)

    hcli_core cli install `hcli_core sample hfm`
    hcli_core cli config hfm core.port 18000
    hcli_core cli config hfm core.auth False

    echo "$(hcli_core cli run hfm) --daemon --log-file=./gunicorn-noauth.log --error-logfile=./gunicorn-noauth-error.log --capture-output" | bash

    cat ./gunicorn-noauth.log
    cat ./gunicorn-noauth-error.log

    huckle cli install http://127.0.0.1:18000
    """

    p1 = subprocess.Popen(['bash', '-c', setup], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p1.communicate()

    if out is not None:
        print(f"STDOUT: {out}")

    if err is not None:
        print(f"STDERR: {err}")

    hello = """
    #!/bin/bash
    set -x

    export HUCKLE_HOME=~/.huckle_test
    eval $(huckle env)
    echo -n '{"hello":"world"}' > hello.json
    cat hello.json | hfm cp -l ./hello.json
    hfm cp -r hello.json > hello1.json
    cat hello1.json
    rm hello.json
    rm hello1.json
    """

    p2 = subprocess.Popen(['bash', '-c', hello], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p2.communicate()
    result = out.decode('utf-8')
    error = err.decode('utf-8')

    if err is not None:
        print(f"STDERR: {err}")

    assert('{"hello":"world"}\n' == result)




import errno
import os
import sys
import shutil
import json
import portalocker

from configparser import ConfigParser
from contextlib import contextmanager
from pathlib import Path
from io import StringIO
from os import listdir
from os.path import isfile, join
from os import path

root = os.path.abspath(os.path.dirname(__file__))
huckle_manpage_path = root + "/data/huckle.1"
home = os.getenv('HUCKLE_HOME') or os.path.expanduser("~")
dot_huckle = "%s/.huckle" % home
dot_huckle_tmp = dot_huckle + "/tmp"
dot_huckle_var = dot_huckle + "/var"
dot_huckle_var_log = dot_huckle_var + "/log"
dot_huckle_scripts = dot_huckle + "/bin"
dot_huckle_config = dot_huckle + "/etc"
dot_huckle_common_config_file_path = dot_huckle_config + "/config"
dot_bash_profile = home + "/.bash_profile"
dot_bashrc = home + "/.bashrc"
credentials_file_path = None
log_file_path = dot_huckle_var_log + "/huckle.log"

# These next variables are dynamically updated from read configuration. Be careful!
url = ""
cliname = "huckle"
cli_manpage_path = dot_huckle + "/tmp"
ssl_verify = "verify"
url_pinning = "dynamic"
credential_helper = "huckle"
auth_mode = "skip"
auth_user_profile = "default"
auth_apikey_profile = "default"

# URL Pinning lookup
pinned_urls = {}

# These hcli_ variables are taken from the semantic types found in the HCLI 1.0 specification
hcli_command_type = "command"
hcli_option_type = "option"
hcli_parameter_type = "parameter"
hcli_execution_type = "execution"

# parses the common huckle configuration to set configured execution
def parse_common_configuration():
    global dot_huckle_common_config_file_path
    common_config_file_path = dot_huckle_common_config_file_path

    parser = ConfigParser()
    parser.read(common_config_file_path)
    if parser.has_section("default"):
        for section_name in parser.sections():
            for name, value in parser.items("default"):
                if name == "log":
                    global log
                    log = value
                if name == "log.level":
                    global log_level
                    log_level = value
                if name == "help":
                    global help_mode
                    help_mode = value
                if name == "protocol.mismatch.timeout":
                    global protocol_mismatch_timeout
                    protocol_mismatch_timeout = value
    else:
        error = f"huckle: no common configuration {common_config_file_path} available."
        raise Exception(error)

# parses the configuration of a given cli to set configured execution
def parse_configuration(cli):
    global credentials_file_path
    credentials_file_path = dot_huckle_config + "/" + cli + "/credentials"

    config_file_path = dot_huckle_config + "/" + cli + "/config"

    parser = ConfigParser()
    parser.read(config_file_path)
    if parser.has_section("default"):
        for section_name in parser.sections():
            for name, value in parser.items("default"):
                if name == "url":
                    global url
                    url = value
                    global cliname
                    cliname = cli
                    global cli_manpage_path
                    cli_manpage_path = cli_manpage_path + "/huckle." + cliname
                if name == "ssl.verify":
                    global ssl_verify
                    ssl_verify = value
                if name == "url.pinning":
                    global url_pinning
                    url_pinning = value
                if name == "credential.helper":
                    global credential_helper
                    credential_helper = value
                if name == "auth.mode":
                    global auth_mode
                    auth_mode = value
                if name == "auth.user.profile":
                    global auth_user_profile
                    auth_user_profile = value
                if name == "auth.apikey.profile":
                    global auth_apikey_profile
                    auth_apikey_profile = value
            if url == "":
                error = f"huckle: no url defined for {cli} under {config_file_path}."
                raise Exception(error)
    else:
        error = f"huckle: no cli configuration {config_file_path} available for {cli}."
        raise Exception(error)

    pinned_file_path = dot_huckle_config + "/" + cli + "/pinned.json"
    try:
        with open(pinned_file_path, 'r') as file:
            global pinned_urls
            pinned_urls = json.load(file)
    except (FileNotFoundError, json.JSONDecodeError):
        pinned_urls = {}

    def success_generator():
        yield ('stdout', b'')

    return success_generator()

def pin_url(command, url, method):
    if command not in pinned_urls:
        pinned_urls[command] = {}
    pinned_urls[command]["url"] = url
    pinned_urls[command]["method"] = method
    save_pinned_urls()

# return a url and method if there's a cache hit.
def get_pinned_url(command):
    if command in pinned_urls:
        return pinned_urls[command]["url"], pinned_urls[command]["method"]
    else:
        return None, None

def save_pinned_urls():
    pinned_file_path = dot_huckle_config + "/" + cliname + "/pinned.json"
    with write_lock(pinned_file_path):
        with open(pinned_file_path, 'w') as file:
            json.dump(pinned_urls, file)

# creates a common configuration file for huckle
def create_common_configuration():
    global dot_huckle_common_config_file_path
    common_config_file_path = dot_huckle_common_config_file_path

    # create the configuration if it doesn't exist
    if not os.path.exists(common_config_file_path):
        create_file(common_config_file_path)
        init_common_configuration()
    else:
        pass

    return

# creates a configuration file for a named cli
def create_configuration(cli, url, description):
    global credentials_file_path
    credentials_file_path = dot_huckle_config + "/" + cli + "/credentials"
    config_file_folder = dot_huckle_config + "/" + cli
    config_file = config_file_folder + "/config"

    try:
        create_folder(config_file_folder)
    except Exception as e:
        error = repr(e)
        logging.error(error)

    if not os.path.exists(config_file):
        try:
            create_file(config_file)
            init_configuration(cli, url, description)
            create_folder(cli_manpage_path + "/huckle." + cli)
            alias_cli(cli)
        except Exception as e:
            error = repr(e)
            logging.error(error)
    else:
        error = f"huckle: the configuration for {cli} already exists. leaving the existing configuration untouched."
        raise Exception(error)

    def success_generator():
        yield ('stdout', f"{cli}".encode('utf-8'))

    return success_generator()

# sets up an alias for a cli so that it can be called directly by name (instead of calling it via the explicit huckle call) 
def alias_cli(cli):
    if not os.path.exists(dot_huckle_scripts + "/" + cli):
        g = open(dot_huckle_scripts + "/" + cli, "a+")
        os.chmod(dot_huckle_scripts + "/" + cli, 0o700)
        g.write("#!/bin/bash\n")
        g.write("huckle cli run " + cli + " \"$@\"")
        g.close

# initializes the common huckle configuration file
def init_common_configuration():
    global dot_huckle_common_config_file_path
    common_config_file_path = dot_huckle_common_config_file_path

    parser = ConfigParser()
    parser.read_file(StringIO(u"[default]"))
    parser.set("default", "log", "skip")
    parser.set("default", "log.level", "info")
    parser.set("default", "help", "text")
    parser.set("default", "protocol.mismatch.timeout", "5")

    with open(common_config_file_path, "w") as config:
        parser.write(config)

# initializes the configuration file of a given cli (initialized when a cli "created")
def init_configuration(cli, url, description):
    global credentials_file_path
    credentials_file_path = dot_huckle_config + "/" + cli + "/credentials"

    config_file_path = dot_huckle_config + "/" + cli + "/config"
    parser = ConfigParser()
    parser.read_file(StringIO(u"[default]"))

    if url is None:
        parser.set("default", "url", "")
    else:
        parser.set("default", "url", url)

    parser.set("default", "description", description)
    parser.set("default", "ssl.verify", "verify")
    parser.set("default", "url.pinning", "dynamic")
    parser.set("default", "credential.helper", "keyring")
    parser.set("default", "auth.mode", "skip")
    parser.set("default", "auth.user.profile", "default")
    parser.set("default", "auth.apikey.profile", "default")

    with open(config_file_path, "w") as config:
        parser.write(config)

# verifies if a given configuration exists
def is_configured(file_path, contains):
    if contains in open(file_path).read():
        return True
    else:
        return False

def get_description(config_path):
    parser = ConfigParser()
    parser.read(config_path)

    description = ""
    if parser.has_section("default"):
        for section_name in parser.sections():
            for name, value in parser.items("default"):
                if name == "description":
                    return value

    return ""

# list all the installed clis
def list_clis():
    def generator():
        try:
            files = [f for f in listdir(dot_huckle_scripts) if isfile(join(dot_huckle_scripts, f))]

            # Handle empty list case first
            if not files:
                yield ('stdout', b'')
                return

            # Find the longest CLI name for padding calculation
            longest_name = max(len(f) for f in files) if files else 0
            # Set column width with exactly 2 spaces after the longest name
            column_width = longest_name + 2

            # Calculate terminal width (fallback to 80 if can't determine)
            terminal_width = 80
            try:
                import shutil
                terminal_width = shutil.get_terminal_size().columns
            except:
                pass  # Use default 80 if we can't get the terminal width

            # Leave some space for the HCLI column and spacing
            desc_max_width = terminal_width - column_width - 2

            # Format and output header
            yield ('stdout', f"{'HCLI':<{column_width}}DESCRIPTION\n".encode('utf-8'))

            # Output all but last item with newlines
            for f in files[:-1]:
                config_path = dot_huckle_config + "/" + f + "/config"
                description = get_description(config_path)

                # Truncate description if needed to fit in terminal
                if desc_max_width > 10 and len(description) > desc_max_width:
                    description = description[:desc_max_width-3] + "..."

                yield ('stdout', f"{f:<{column_width}}{description}\n".encode('utf-8'))

            # Last item without newline
            if files:
                f = files[-1]
                config_path = dot_huckle_config + "/" + f + "/config"
                description = get_description(config_path)

                # Truncate description if needed to fit in terminal
                if desc_max_width > 10 and len(description) > desc_max_width:
                    description = description[:desc_max_width-3] + "..."

                yield ('stdout', f"{f:<{column_width}}{description}".encode('utf-8'))

        except Exception as e:
            error = f"huckle: error listing clis: {str(e)}"
            raise Exception(error)

    return generator()

# remove a cli
def remove_cli(cli):
    script_path = dot_huckle_scripts + "/" + cli
    config_path = dot_huckle_config + "/" + cli

    def generator():
        if path.exists(script_path):
            os.remove(script_path)
            shutil.rmtree(config_path)
            yield ('stdout', b'')
        elif path.exists(config_path):
            shutil.rmtree(config_path)
            error = f"huckle: {cli} was not installed, however some configuration remained and is now cleanly deleted."
            raise Exception(error)
        else:
            error = f"huckle: {cli} is not installed."
            raise Exception(error)

    return generator()

# remove a pinned url cache
def flush_pinned_urls(cli):
    pinned_file_path = dot_huckle_config + "/" + cli + "/pinned.json"

    def generator():
        if path.exists(pinned_file_path):
            try:
                os.remove(pinned_file_path)
                yield ('stdout', b'')
            except Exception as e:
                error = f"huckle: error flushing pinned urls: {str(e)}"
                raise Exception(error)
        else:
            error = f"huckle: no pinned url cache to flush for {cli}."
            raise Exception(error)

    return generator()

# lists all the configuration parameters of a cli
def config_list(cli):
    config_file_path = dot_huckle_config + "/" + cli + "/config"
    parser = ConfigParser()
    parser.read(config_file_path)

    def generator():
        try:
            # Get all content first
            all_lines = []
            for section_name in parser.sections():
                all_lines.append(f"[{section_name}]")
                for name, value in parser.items(section_name):
                    all_lines.append(f'{name} = {value}')

            # Handle empty case
            if not all_lines:
                yield ('stdout', b'')
                return

            # Output all but last line with newlines
            for line in all_lines[:-1]:
                yield ('stdout', f"{line}\n".encode('utf-8'))

            # Output last line without newline
            if all_lines:
                yield ('stdout', all_lines[-1].encode('utf-8'))

        except Exception as error:
            error = "huckle: unable to list configuration."
            raise Exception(error)

    return generator()

# update a configured parameter to a new value
def update_parameter(cli, parameter, value):
    config_file_path = dot_huckle_config + "/" + cli + "/config"
    parser = ConfigParser()
    parser.read(config_file_path)

    def generator():
        try:
            parser.set('default', parameter, value)
            with write_lock(config_file_path):
                with open(config_file_path, "w") as config:
                    parser.write(config)
            yield ('stdout', b'')
        except Exception as error:
            error = "huckle: unable to update configuration."
            raise Exception(error)

    return generator()

# unset a configured parameter
def unset_parameter(cli, parameter):
    config_file_path = dot_huckle_config + "/" + cli + "/config"
    parser = ConfigParser()
    parser.read(config_file_path)

    def generator():
        try:
            if parser.has_section('default') and parser.has_option('default', parameter):
                parser.remove_option('default', parameter)
            with write_lock(config_file_path):
                with open(config_file_path, "w") as config:
                    parser.write(config)
            yield ('stdout', b'')
        except Exception as error:
            error = "huckle: unable to update configuration."
            raise Exception(error)

    return generator()

# get a configured parameter
def get_parameter(cli, parameter):
    config_file_path = dot_huckle_config + "/" + cli + "/config"
    parser = ConfigParser()
    parser.read(config_file_path)

    def generator():
        try:
            value = parser.get('default', parameter)
            yield ('stdout', value.encode('utf-8'))
        except Exception as error:
            error = "huckle: unable to retrieve configuration."
            raise Exception(error)

    return generator()

# creates a folder at "path"
def create_folder(path):

    # Skip lock only if this is the actual user's home directory (not a custom HUCKLE_HOME)
    if path == os.path.expanduser("~"):
        if not os.path.exists(path):
            os.makedirs(path)
        return

    with write_lock(path):
        if not os.path.exists(path):
            os.makedirs(path)

# creates a file at "path"
def create_file(path):
    with write_lock(path):
        if not os.path.exists(path):
            flags = os.O_CREAT | os.O_EXCL | os.O_WRONLY

            try:
                file_handle = os.open(path, flags, 0o0600)
            except OSError as e:
                if e.errno == errno.EEXIST:  # Failed since the file already exists.
                    pass
                else:
                    raise
            else:
                with os.fdopen(file_handle, 'w') as file_obj:
                    file_obj.write("")
                    file_obj.close

@contextmanager
def write_lock(file_path):
    lockfile = Path(file_path).with_suffix('.lock')
    with portalocker.Lock(lockfile, timeout=10) as lock:
        yield

        # we cleanup the lock if successful.
        try:
            if lockfile.exists():
                os.unlink(lockfile)
        except OSError:
            pass

create_folder(home)
create_folder(dot_huckle)
create_folder(dot_huckle_tmp)
create_folder(dot_huckle_config)
create_folder(dot_huckle_var)
create_folder(dot_huckle_var_log)
create_folder(dot_huckle_scripts)
create_file(dot_bash_profile)

# create and load the common huckle configuration for logging before first log initialization
create_common_configuration()
parse_common_configuration()

# we load the logger after the configuration is in otherwise we get no logger
from huckle import logger

# Map string log levels to logger constants
LOG_LEVELS = {
    'debug': logger.DEBUG,
    'info': logger.INFO,
    'warning': logger.WARNING,
    'error': logger.ERROR,
    'critical' : logger.CRITICAL
}

log_level = LOG_LEVELS.get(log_level.lower(), logger.INFO)
logging = logger.Logger(log=log)
logging.setLevel(log_level)



import subprocess
import os
import pytest

def test_hco_key_admin_keyring_credential_helper(gunicorn_server_auth, cleanup):
    hello = """
    #!/bin/bash
    set -x

    export HUCKLE_HOME=~/.huckle_test
    eval $(huckle env)

    huckle cli config hco credential.helper keyring
    hco key admin

    """

    p2 = subprocess.Popen(['bash', '-c', hello], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = p2.communicate()
    result = out.decode('utf-8')

    key_id, api_key, timestamp = result.split()

    # Expected lengths based on your example
    assert len(key_id) == 10, f"Key ID length should be 10, got {len(key_id)}"
    assert len(api_key) == 92, f"API key length should be 107, got {len(api_key)}"
    assert len(timestamp) == 32, f"Timestamp length should be 32, got {len(timestamp)}"

    # Verify format patterns
    assert key_id.isalnum(), "Key ID should be alphanumeric"
    assert api_key.startswith("hcoak_"), "API key should start with 'hcoak_'"

def test_hco_ls_keyring_credential_helper(gunicorn_server_auth, cleanup):
    hello = """
    #!/bin/bash
    set -x

    export HUCKLE_HOME=~/.huckle_test
    eval $(huckle env)

    huckle cli config hco credential.helper keyring
    hco ls

    """

    p2 = subprocess.Popen(['bash', '-c', hello], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = p2.communicate()
    result = out.decode('utf-8')

    assert('admin' in result)

def test_jsonf_keyring_credential_helper(gunicorn_server_auth, cleanup):
    hello = """
    #!/bin/bash

    export HUCKLE_HOME=~/.huckle_test
    eval $(huckle env)

    huckle cli config jsonf credential.helper keyring
    echo '{"hello":"world"}' | jsonf go
    """

    p2 = subprocess.Popen(['bash', '-c', hello], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = p2.communicate()
    result = out.decode('utf-8')

    assert('{\n    "hello": "world"\n}' in result)

def test_hco_key_admin_huckle_credential_helper(gunicorn_server_auth, cleanup):
    hello = """
    #!/bin/bash
    set -x

    export HUCKLE_HOME=~/.huckle_test
    eval $(huckle env)

    huckle cli config hco credential.helper huckle
    hco key admin

    """

    p2 = subprocess.Popen(['bash', '-c', hello], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = p2.communicate()
    result = out.decode('utf-8')

    key_id, api_key, timestamp = result.split()

    # Expected lengths based on your example
    assert len(key_id) == 10, f"Key ID length should be 10, got {len(key_id)}"
    assert len(api_key) == 92, f"API key length should be 107, got {len(api_key)}"
    assert len(timestamp) == 32, f"Timestamp length should be 32, got {len(timestamp)}"

    # Verify format patterns
    assert key_id.isalnum(), "Key ID should be alphanumeric"
    assert api_key.startswith("hcoak_"), "API key should start with 'hcoak_'"

def test_hco_ls_huckle_credential_helper(gunicorn_server_auth, cleanup):
    hello = """
    #!/bin/bash
    set -x

    export HUCKLE_HOME=~/.huckle_test
    eval $(huckle env)

    huckle cli config hco credential.helper huckle
    hco ls

    """

    p2 = subprocess.Popen(['bash', '-c', hello], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = p2.communicate()
    result = out.decode('utf-8')

    assert('admin' in result)

def test_jsonf_huckle_credential_helper(gunicorn_server_auth, cleanup):
    hello = """
    #!/bin/bash

    export HUCKLE_HOME=~/.huckle_test
    eval $(huckle env)

    huckle cli config jsonf credential.helper huckle
    echo '{"hello":"world"}' | jsonf go
    kill $(ps aux | grep '[g]unicorn' | awk '{print $2}')
    """

    p2 = subprocess.Popen(['bash', '-c', hello], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    out, err = p2.communicate()
    result = out.decode('utf-8')

    assert('{\n    "hello": "world"\n}' in result)



from huckle import config
from huckle.auth import credential

import sys
import shlex
import io
from subprocess import call

from huckle import package
from huckle import hclinav
from huckle import logger

from contextlib import contextmanager

logging = logger.Logger()


# navigate through the command line sequence for a given cliname
def navigate(argv):
    nav = hclinav.navigator(root=config.url, apiname=config.cliname)

    # we try to fail fast if the service isn't reachable
    try:
        nav()["name"]
    except Exception as e:
        error = config.cliname + ": unable to navigate HCLI 1.0 compliant semantics. wrong url, or the service isn't running? " + str(nav.uri)
        raise Exception(error)

    # if we're configured for url pinning, we try to get a cache hit
    if config.url_pinning == "pin":
        command = reconstruct_command(argv)
        url, method = config.get_pinned_url(command)
        if url:
            logging.debug("found: [" + command + "] " + url + " " + method)
            return hclinav.flexible_executor(url, method)

    if len(argv) == 1:
        return hclinav.traverse_execution(nav)

    length = len(argv[1:])
    for i, x in enumerate(argv[1:]):
        nav = hclinav.traverse_argument(nav, x)

        if type(nav) == tuple:
            def generator():
                yield nav
            return generator()
        if i == length - 1:
            return hclinav.traverse_execution(nav)

@contextmanager
def stdin(stream):
    if isinstance(stream, (str, bytes)):
        stream = io.BytesIO(stream.encode() if isinstance(stream, str) else stream)
    elif not hasattr(stream, 'read'):
        stream = io.BytesIO(b''.join(stream))

    sys.stdin = stream # Redirect sys.stdin to a provided io stream (e.g. io.BytesIO)

    try:
        yield

    finally:
        sys.stdin = sys.__stdin__

# Execute CLI commands either from command line or as a library.
# Now checks if we're already in a stdin context.
def cli(commands=None):
    if commands is not None:
        needs_stdin = not isinstance(sys.stdin, io.BytesIO)

        if needs_stdin:
            with stdin(io.BytesIO(b'')):
                return __execute_cli(commands)
        else:
            return __execute_cli(commands)
    else:
        return __execute_cli(None)

# huckle's minimal set of commands
def __execute_cli(commands=None):
    if commands is not None:
        argv = shlex.split(commands)
    else:
        argv = sys.argv
        argv[0] = "huckle"

    if argv[0] == "huckle":
        pass
    else:
        config.parse_configuration(argv[0])
        return navigate(argv[0:])

    if len(argv) > 2:

        if argv[1] == "cli" and argv[2] == "run":

            if len(argv) > 3:
                config.parse_configuration(argv[3])
                return navigate(argv[3:])

            else:
                return huckle_help()

        elif argv[1] == "cli" and argv[2] == "install":

            if len(argv) > 3:
                return hclinav.install(argv[3])

            else:
                return huckle_help()

        elif argv[1] == "cli" and argv[2] == "ls":
            return config.list_clis()

        elif argv[1] == "cli" and argv[2] == "rm":
            if len(argv) > 3:
                return config.remove_cli(argv[3])

            else:
                return huckle_help()

        elif argv[1] == "cli" and argv[2] == "flush":
            if len(argv) > 3:
                return config.flush_pinned_urls(argv[3])

            else:
                return huckle_help()

        elif argv[1] == "cli" and argv[2] == "config":
            if len(argv) == 4:
                return config.config_list(argv[3])
            elif len(argv) == 5:
                return config.get_parameter(argv[3], argv[4])
            elif len(argv) == 6:
                if sys.argv[4] == "--unset":
                    return config.unset_parameter(sys.argv[3], sys.argv[5])
                else:
                    return config.update_parameter(sys.argv[3], sys.argv[4], sys.argv[5])

            else:
                return huckle_help()

        elif argv[1] == "cli" and argv[2] == "credential":
            if len(argv) == 5:
                if not sys.stdin.isatty():
                    secret = sys.stdin.read()

                    if isinstance(secret, bytes):
                        secret = secret.decode().strip()
                    else:
                        secret = secret.strip()

                    config.parse_configuration(argv[3])
                    credentials = credential.CredentialManager()
                    return credentials.update_credential(argv[4], secret)
                else:
                    return huckle_help()
            else:
                return huckle_help()

        elif argv[1] == "help":
            if config.help_mode == "text":
                with open(config.huckle_manpage_path, "r") as f:

                    text = f.read()
                    def generator():
                        yield ('stdout', hclinav.troff_to_text(text).encode('utf-8'))

                    return generator()
                f.close()
            elif config.help_mode == "man":
                call(["man", config.huckle_manpage_path])
                sys.exit(0)

        else:
            return huckle_help()

    elif len(argv) == 2:

        if argv[1] == "--version":
            return show_dependencies()

        elif argv[1] == "env":
            text = "export PATH=$PATH:" + config.dot_huckle_scripts + "\n\n"
            text += "# To point your shell to huckle's HCLI entrypoint scripts, run:\n"
            text += "# eval $(huckle env)"

            def generator():
                yield ('stdout', text.encode('utf-8'))

            return generator()

        elif argv[1] == "help":
            if config.help_mode == "text":
                with open(config.huckle_manpage_path, "r") as f:

                    text = f.read()
                    def generator():
                        yield ('stdout', hclinav.troff_to_text(text).encode('utf-8'))

                    return generator()
                f.close()
            elif config.help_mode == "man":
                call(["man", config.huckle_manpage_path])
                sys.exit(0)

        else:
            return huckle_help()

    else:
        return huckle_help()

def huckle_help():
    error = "for help, use:\n\n  huckle help"
    raise Exception(error)

# show huckle's version and the version of its dependencies

def show_dependencies():
    def parse_dependency(dep_string):
        # Common version specifiers
        specifiers = ['==', '>=', '<=', '~=', '>', '<', '!=']

        # Find the first matching specifier
        for specifier in specifiers:
            if specifier in dep_string:
                name, version = dep_string.split(specifier, 1)
                return name.strip(), specifier, version.strip()

        # If no specifier found, return just the name
        return dep_string.strip(), '', ''

    dependencies = ""
    for dep in package.dependencies:
        name, specifier, version = parse_dependency(dep)
        if version:  # Only add separator if there's a version
            dependencies += f" {name}/{version}"
        else:
            dependencies += f" {name}"

    def generator():
        yield ('stdout', f"huckle/{package.__version__}{dependencies}".encode('utf-8'))

    return generator()

def reconstruct_command(args):
    reconstructed = []
    for arg in args:
        # escape every single quote
        arg = arg.replace("'", "\\'")
        reconstructed.append(arg)
    return ' '.join(reconstructed)



from subprocess import call
from restnavigator import Navigator
from functools import partial
from urllib.parse import urlparse, parse_qs, unquote

# avoid broken pipe signal crashing the program
from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE,SIG_DFL)

# huckle's imports
from huckle import config
from huckle import logger
from huckle.auth import credential
from huckle.auth import authenticator

import sys
import os
import fcntl
import json
import subprocess
import time
import requests
import errno
import socket
import certifi
import textwrap
import re
import shutil

try:
        from urllib import quote  # Python 2.X
except ImportError:
        from urllib.parse import quote  # Python 3+

logging = logger.Logger()


# produces a navigator that starts navigating from the root and with an api display name of apiname
def navigator(root, apiname):
    s = requests.Session()

    # ssl verify configuration check
    if config.ssl_verify == "verify":
        s.verify = certifi.where()
    elif config.ssl_verify == "skip":
        logging.warning("SSL verification is being skipped. This will leak credentials on the network if using authentication.")
        import warnings
        from urllib3.exceptions import InsecureRequestWarning
        warnings.simplefilter('ignore', InsecureRequestWarning)
        s.verify = False

#     if config.auth_mode == "basic":
#         logging.debug("HTTP Basic Authentication...")
#         credentials = credential.CredentialManager()
#         s.auth = requests.auth.HTTPBasicAuth(*(credentials.find()))
# 
#     elif config.auth_mode == "hcoak":
#         logging.debug("HCLI Core API Key Authentication...")
#         credentials = credential.CredentialManager()
#         s.auth = authenticator.HCOAKBearerAuth(*(credentials.hcoak_find()))

#     # Ensure root URL has proper scheme
#     if not root.lower().startswith(('http://', 'https://')):
#         # Add default http:// if no scheme is specified
#         root = 'http://' + root
#         logging.debug(f"huckle: no scheme specified, using: {root}")

    # For HTTPS URLs, add a fixed timeout to prevent indefinite stalling
    if root.lower().startswith('https://'):
        parsed_url = urlparse(root)
        hostname = parsed_url.netloc.split(':')[0]
        port = parsed_url.port or 443

        try:
            # Make a simple HEAD request with a short timeout
            # This will quickly reveal if HTTPS is working
            response = s.head(root, timeout=(5,5))
            # If we get here, HTTPS is working fine
        except requests.exceptions.SSLError:
            # SSL errors mean the server is responding but has certificate issues
            # This is not a protocol mismatch, so continue
            pass
        except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as e:
            # Connection errors or timeouts might indicate protocol mismatch
            # If it's "connection reset" or similar, server is responding somehow
            error_str = str(e).lower()
            if "timed out" in error_str or "connection aborted" in error_str:
                # These specific errors often indicate protocol mismatch
                logging.warning("huckle: https connection appears to be stalled - possible protocol mismatch")
                http_url = root.replace('https://', 'http://', 1)
                error_msg = (f"{config.cliname}: https connection attempt stalled. "
                            f"The server at {hostname} may not support https. "
                            f"Try using {http_url}")
                raise Exception(error_msg)

    nav = Navigator.hal(root=root, apiname=apiname, session=s)

    return nav

# attempts to traverse through an hcli document with a command line argument
def traverse_argument(nav, arg):
    ilength = 0
    try:
        ilength = len(nav.links()["cli"])
    except Exception as warning:
        error = config.cliname + ": unable to navigate HCLI 1.0 compliant semantics. wrong url, or the service isn't running? " + str(nav.uri)
        raise Exception(error)

    for j, y in enumerate(nav.links()["cli"]):

        # we give first precedence to help so that help is easily accessible at all time.
        if arg == "help":
            if config.help_mode == "text":
                (content, path) = hcli_to_troff(nav, save_to_file=False)
                text = troff_to_text(content)
                return  ('stdout', text.encode('utf-8'))
            elif config.help_mode == "man":
                (content, path) = hcli_to_troff(nav, save_to_file=True)
                call(["man", path])
                sys.exit(0)

        # we give precedence to parameter traversal to help avoid forcing double quoting on the command line
        try:
            for k, l in enumerate(nav.links()["cli"][j]):
                hcli_type = l.links()["profile"][0].uri.split('#', 1)[1]
                if hcli_type == config.hcli_parameter_type:
                    if not (arg.startswith('\"') and arg.endswith('\"')):
                        arg = '\"' + arg + '\"'
                    nav = l["cli"][0](hcli_param=quote(arg))
                    return nav
        except:
            pass

        tempnav = nav.links()["cli"][j]
        try:
            if tempnav()["name"] == arg:
                nav = tempnav["cli"][0]
                return nav
        except:
            error = config.cliname + ": " + arg + ": " + "command not found.\n"
            error += for_help()
            raise Exception(error)

        if j == ilength - 1:
            error = config.cliname + ": " + arg + ": " + "command not found.\n"
            error += for_help()
            raise Exception(error)

# attempts to traverse through an execution. (only attempted when we've run out of command line arguments to parse)
def traverse_execution(nav):
    try:
        for k, z in enumerate(nav.links()["cli"]):
            tempnav = nav.links()["cli"][k]

            hcli_type = tempnav.links()["profile"][0].uri.split('#', 1)[1]
            if hcli_type == config.hcli_execution_type:
                method = tempnav()["http"]
                nav = tempnav["cli"][0]
                return flexible_executor(nav.uri, method)
    except KeyError:
        error = config.cliname + ": " + "command/parameter confusion. try escaping parameter: e.g., \\\"param\\\" or \\\'param\\\'.\n"
        error += for_help()
        raise Exception(error)

    error = config.cliname + ": " + "unable to execute.\n"
    error += for_help()
    raise Exception(error)

# attempts to pull at the root of the hcli to auto configure the cli
def install(url):
    nav = navigator(root=url, apiname="unknown")
    version = None
    try:
        version = nav()["hcli_version"]
    except Exception as warning:
        pass

    if version == "1.0":
        cli = nav()["name"]

        # Extract description from the section where name is "name"
        description = ""
        for section in nav().get("section", []):
            if section.get("name") == "name":
                description = section.get("description")
                break

        yield from config.create_configuration(cli, url, description)
    else:
        configurations = []
        for k, z in enumerate(nav.links()["cli"]):
            yield from install(nav.links()["cli"][k].uri)

            # Add separator after each HCLI except the last one
            if k < len(nav.links()["cli"]) - 1:
                yield ('stdout', b'\n')

# generates an OPTIONS and COMMANDS section to add to a text page
def options_and_commands_to_text(navigator):

    # This block outputs an OPTIONS section, in the man page, alongside each available option flag and its description
    options = ""
    option_count = 0
    for i, x in enumerate(navigator.links()["cli"]):
        tempnav = navigator.links()["cli"][i]
        hcli_type = tempnav.links()["profile"][0].uri.split('#', 1)[1]
        if hcli_type == config.hcli_option_type:
            option_count += 1
            options = options + "       " + tempnav()["name"] + "\n"
            options = options + "              " + tempnav()["description"] + "\n"
    if option_count > 0:
        options = "OPTIONS\n" + options

    # This block outputs a COMMANDS section, in the man page, alongside each available command and its description
    commands = ""
    command_count = 0
    for i, x in enumerate(navigator.links()["cli"]):
        tempnav = navigator.links()["cli"][i]
        hcli_type = tempnav.links()["profile"][0].uri.split('#', 1)[1]
        if hcli_type == config.hcli_command_type:
            command_count += 1
            commands = commands + "       " + tempnav()["name"] + "\n"
            commands = commands + "              " + tempnav()["description"] + "\n"
    if command_count > 0:
        commands = "COMMANDS\n" + commands

    return options + commands

# format hcli navigation to manpage like format
def hcli_to_troff(navigator, save_to_file=True):
    man_content = ""

    # Add the title header
    man_content += ".TH " + navigator()["name"].upper() + " 1 \n"

    # Process each section
    for i, x in enumerate(navigator()["section"]):
        section = navigator()["section"][i]
        man_content += ".SH " + section["name"].upper() + "\n"
        man_content += section["description"].replace("\\n\\n", "\n.sp\n") + "\n"

    # Add options and commands
    man_content += options_and_commands_to_man(navigator)

    # Save to file if requested
    file_path = None
    if save_to_file:
        millis = str(time.time())
        dynamic_doc_path = config.cli_manpage_path + "/" + config.cliname + "." + millis + ".man"
        config.create_folder(config.cli_manpage_path)
        config.create_file(dynamic_doc_path)

        with open(dynamic_doc_path, "w") as f:
            f.write(man_content)

        file_path = dynamic_doc_path

    return (man_content, file_path)

def man_to_txt(man_page):
    try:
        process = subprocess.run(['man', man_page], capture_output=True, text=True, check=True)
        process_col = subprocess.run(['col', '-b'], input=process.stdout, capture_output=True, text=True, check=True)
        return process_col.stdout
    except subprocess.CalledProcessError as e:
        error = "unable to convert help man page to text"
        raise Exception(error)

# generates an OPTIONS and COMMANDS section to add to a man page
def options_and_commands_to_man(navigator):
    # This block outputs an OPTIONS section, in the man page, alongside each available option flag and its description
    options = ""
    option_count = 0
    for i, x in enumerate(navigator.links()["cli"]):
        tempnav = navigator.links()["cli"][i]
        hcli_type = tempnav.links()["profile"][0].uri.split('#', 1)[1]
        if hcli_type == config.hcli_option_type:
            option_count += 1
            options = options + ".IP " + tempnav()["name"] + "\n"
            options = options + tempnav()["description"].replace("\\n\\n", "\n.sp\n") + "\n"
    if option_count > 0:
        options = ".SH OPTIONS\n" + options

    # This block outputs a COMMANDS section, in the man page, alongside each available command and its description
    commands = ""
    command_count = 0
    for i, x in enumerate(navigator.links()["cli"]):
        tempnav = navigator.links()["cli"][i]
        hcli_type = tempnav.links()["profile"][0].uri.split('#', 1)[1]
        if hcli_type == config.hcli_command_type:
            command_count += 1
            commands = commands + ".IP " + tempnav()["name"] + "\n"
            commands = commands + tempnav()["description"].replace("\\n\\n", "\n.sp\n") + "\n"
    if command_count > 0:
        commands = ".SH COMMANDS\n" + commands

    return options + commands

# standard error message to tell users to go check the help pages (man pages)
def for_help():
    text = ""
    text += "for help, use:\n\n"
    text += "  " + config.cliname + " help\n"
    text += "  " + config.cliname + " <command> help"
    return text

# a flexible executor that can work with the application/octet-stream media-type (per HCLI 1.0 spec)
def flexible_executor(url, method):
    # we take into account how a CLI should interact with SSL verification
    ssl_verify = "verify"
    if config.ssl_verify == "verify":
        ssl_verify = certifi.where()
    elif config.ssl_verify == "skip":
        ssl_verify = False

    # if we're configured to pin final urls, we setup the cache for future hits
    if config.url_pinning == "pin":
        parsed_url = urlparse(url)
        params = parse_qs(parsed_url.query)
        command_encoded = params.get('command', [None])[0]  # Returns a list, so take the first item
        if command_encoded:
            final_command = unquote(unquote(command_encoded))

            # Replace characters to match original format and pin the url
            final_command = final_command.replace('"', '').replace("'", r"\'")
            config.pin_url(final_command, url, method)
            logging.debug("pinned: [" + final_command + "] " + url + " " + method)

    auth_mode = None
    if config.auth_mode == "basic":
        credentials = credential.CredentialManager()
        auth_mode = requests.auth.HTTPBasicAuth(*(credentials.find()))
    elif config.auth_mode == "hcoak":
        credentials = credential.CredentialManager()
        auth_mode = authenticator.HCOAKBearerAuth(*(credentials.hcoak_find()))

    if method == "get":
        r = requests.get(url, stream=True, verify=ssl_verify, auth=auth_mode)
        return output_chunks(r)
    if method == "post":
        if not sys.stdin.isatty():

            logging.debug("Attempting to stream POST data...")
            headers = {'content-type': 'application/octet-stream'}
            stream = nbstdin()

            r = requests.post(url, data=stream.read(), headers=headers, stream=True, verify=ssl_verify, auth=auth_mode)
            return output_chunks(r)
        else:
            logging.debug("No data to stream. Empty POST data.")
            r = requests.post(url, data=None, stream=True, verify=ssl_verify, auth=auth_mode)
            return output_chunks(r)

    return

# outputs the response received from an execution
def output_chunks(response):
    if response.status_code >= 400:
        f = getattr(sys.stderr, 'buffer', sys.stderr)
        if response.headers['content-type'] == 'application/problem+json':
            try:
                problem_detail = json.loads(response.content)
                logging.error(problem_detail)
                error_msg = f"{problem_detail.get('detail', '')}"
                yield ('stderr', f"{config.cliname}: {error_msg}".encode('utf-8'))
            except (json.JSONDecodeError, KeyError) as error:
                raise Exception(error)
        else:
            for chunk in response.iter_content(16384):
                if chunk:
                    yield ('stderr', chunk)
    else:
        f = getattr(sys.stdout, 'buffer', sys.stdout)
        for chunk in response.iter_content(16384):
            if chunk:
                yield ('stdout', chunk)

# wraps stdin into a unbuffered generator
class nbstdin:
    def __init__(self):
        None

    def read(self):
        try:
            f = os.fdopen(sys.stdin.fileno(), 'rb', 0)
            logging.debug("huckle: It's likely a real command line environment stdin inputstream.")
            with f as fis:
                for chunk in iter(partial(fis.read, 16384), b''):
                    yield chunk
        except Exception as e:
            logging.debug("huckle: It's likely a huckle library use context. Falling back to an assumed doctored io.BytesIO inputstream.")
            for chunk in iter(partial(sys.stdin.read, 16384), b''):
                yield chunk

def troff_to_text(content, width=None):
    # If width is not specified, try to get the terminal size
    if width is None:
        try:
            columns, _ = shutil.get_terminal_size()
            width = columns
        except Exception:
            # Fall back to 80 if we can't determine terminal size
            width = 80

    # Helper function to handle troff escape characters
    def process_escapes(text):
        # Generic rule: remove backslash before any character
        text = re.sub(r'\\(.)', r'\1', text)
        return text

    # Extract the man page title from .TH line
    title_match = re.search(r'\.TH\s+(\S+)\s+(\S+)', content)
    if title_match:
        name = title_match.group(1)
        section = title_match.group(2)
        name_section = f"{name}({section})"
        centered_text = "User Commands"

        # Calculate proper alignment positions for header
        left_text = name_section
        center_text = centered_text
        right_text = name_section

        # Create properly aligned header
        left_part = left_text
        center_start = (width - len(center_text)) // 2
        center_part = " " * (center_start - len(left_part)) + center_text
        right_start = width - len(right_text)
        right_part = " " * (right_start - len(left_part) - len(center_part)) + right_text

        header = left_part + center_part + right_part

        # Create right-aligned footer
        footer = " " * (width - len(name_section)) + name_section
    else:
        header = ""
        footer = ""

    # Initialize result with header
    result = [header, ""] if header else []

    # Process the content line by line
    lines = content.split('\n')
    i = 0
    while i < len(lines):
        line = lines[i].strip()

        if line.startswith('.B'):
            # For top-level .B directives, collect them as part of the next regular text
            bold_text = line[2:].strip()
            i += 1
            continue

        # Process .SH (section header)
        if line.startswith('.SH'):
            # Add only a single blank line before section header
            if result and result[-1] != "":
                result.append("")
            section_name = process_escapes(line[4:].strip().strip('"'))
            result.append(section_name)
            i += 1

            # Process the content until the next .SH or end
            section_content = []
            paragraph_lines = []
            is_first_ip = True  # Flag to track first .IP in section

            while i < len(lines):
                current = lines[i].strip()

                # Check for next section header
                if current.startswith('.SH'):
                    break

                if current.startswith('.B'):
                    bold_text = current[2:].strip()
                    if bold_text:
                        paragraph_lines.append(bold_text)
                    i += 1
                    continue

                # Process subsection header (.SS)
                if current.startswith('.SS'):
                    if paragraph_lines:
                        para_text = ' '.join(paragraph_lines)
                        wrapped_lines = textwrap.wrap(para_text, width=width-7)
                        for wrapped_line in wrapped_lines:
                            result.append(f"       {wrapped_line}")
                        result.append("")
                        paragraph_lines = []

                    if result and result[-1] != "":
                        result.append("")
                    subsection_name = process_escapes(current[4:].strip().strip('"'))
                    result.append(f"   {subsection_name}")
                    i += 1
                    is_first_ip = True  # Reset flag for new subsection
                    continue

                # Process indented paragraph (.IP)
                if current.startswith('.IP'):
                    if paragraph_lines:
                        para_text = ' '.join(paragraph_lines)
                        wrapped_lines = textwrap.wrap(para_text, width=width-7)
                        for wrapped_line in wrapped_lines:
                            result.append(f"       {wrapped_line}")
                        paragraph_lines = []

                    # Add blank line before .IP entry only if it's not the first .IP
                    if not is_first_ip:
                        result.append("")

                    is_first_ip = False  # Update flag after processing first .IP

                    item_match = re.search(r'\.IP\s+"([^"]+)"', current)
                    if item_match:
                        item_name = process_escapes(item_match.group(1))
                    else:
                        item_name = process_escapes(current[3:].strip().strip('"'))

                    result.append(f"       {item_name}")
                    i += 1

                    desc_text = []
                    # Check for .B immediately following .IP
                    if i < len(lines) and lines[i].strip().startswith('.B'):
                        bold_text = process_escapes(lines[i].strip()[2:].strip())
                        if bold_text:
                            desc_text.append(bold_text)
                        i += 1

                    while i < len(lines) and not (lines[i].strip().startswith('.') and 
                                               not lines[i].strip().startswith('.br') and 
                                               not lines[i].strip().startswith('.sp') and
                                               not lines[i].strip().startswith('.B')):
                        if lines[i].strip().startswith('.sp'):
                            if desc_text:
                                wrapped_desc = textwrap.wrap(' '.join(desc_text), width=width-14)
                                for wrapped_line in wrapped_desc:
                                    result.append(f"              {wrapped_line}")
                                result.append("")
                                desc_text = []
                        elif lines[i].strip().startswith('.br'):
                            if desc_text:
                                wrapped_desc = textwrap.wrap(' '.join(desc_text), width=width-14)
                                for wrapped_line in wrapped_desc:
                                    result.append(f"              {wrapped_line}")
                                desc_text = []
                        elif lines[i].strip().startswith('.B'):
                            if desc_text:
                                wrapped_desc = textwrap.wrap(' '.join(desc_text), width=width-14)
                                for wrapped_line in wrapped_desc:
                                    result.append(f"              {wrapped_line}")
                                desc_text = []
                        else:
                            if not lines[i].strip().startswith('.'):
                                desc_text.append(lines[i].strip())
                        i += 1

                    if desc_text:
                        wrapped_desc = textwrap.wrap(' '.join(desc_text), width=width-14)
                        for wrapped_line in wrapped_desc:
                            result.append(f"              {wrapped_line}")

                    continue

                if current.startswith('.sp'):
                    if paragraph_lines:
                        para_text = ' '.join(paragraph_lines)
                        wrapped_lines = textwrap.wrap(para_text, width=width-7)
                        for wrapped_line in wrapped_lines:
                            result.append(f"       {wrapped_line}")
                        result.append("")
                        paragraph_lines = []
                    i += 1
                    continue

                if current.startswith('.br'):
                    if paragraph_lines:
                        para_text = ' '.join(paragraph_lines)
                        wrapped_lines = textwrap.wrap(para_text, width=width-7)
                        for wrapped_line in wrapped_lines:
                            result.append(f"       {wrapped_line}")
                        paragraph_lines = []
                    i += 1
                    continue

                if not current.startswith('.'):
                    processed_text = process_escapes(current)
                    paragraph_lines.append(processed_text)

                i += 1

            if paragraph_lines:
                para_text = ' '.join(paragraph_lines)
                wrapped_lines = textwrap.wrap(para_text, width=width-7)
                for wrapped_line in wrapped_lines:
                    result.append(f"       {wrapped_line}")

            continue

        i += 1

    # Add footer with empty line before it
    if footer:
        result.append("")
        result.append(footer)

    return '\n'.join(result)



import sys
import os
import keyring

from configparser import ConfigParser
from contextlib import suppress

from huckle import logger
from huckle import config

log = logger.Logger()


class CredentialManager:
    def __init__(self):
        self._credentials = None
        self.credentials_file_path = config.credentials_file_path

    @property
    def credentials(self):
        return self._credentials

    def parse_credentials(self):
        try:
            log.debug("keyring backend: " + str(keyring.get_keyring()))

            with open(self.credentials_file_path, 'r') as cred_file:
                parser = ConfigParser(interpolation=None)
                log.debug("Loading credentials")
                log.debug(self.credentials_file_path)
                parser.read_file(cred_file)

                # Check for unique usernames across all sections
                usernames = set()
                for section in parser.sections():
                    if parser.has_option(section, "username"):
                        username = parser.get(section, "username")
                        if username in usernames:
                            error = f"duplicate username '{username}' found in {self.credentials_file_path}."
                            self._credentials = None
                            raise Exception(error)
                        usernames.add(username)

                new_credentials = {}
                for section_name in parser.sections():
                    new_credentials[str(section_name)] = []
                    for name, value in parser.items(section_name):
                        if config.credential_helper == 'keyring':
                            if name == 'password' or name == 'keyid':
                                pass

                            # For keyring mode, get passwords/apikeys from keyring
                            if name == 'username':
                                new_credentials[str(section_name)].append({str(name): str(value)})
                                password = keyring.get_password(config.url, value)
                                new_credentials[str(section_name)].append({'password': str(password) if password is not None else ''})
                            elif name == 'keyid':
                                new_credentials[str(section_name)].append({str(name): str(value)})
                                apikey = keyring.get_password(config.url, value)
                                new_credentials[str(section_name)].append({'apikey': str(apikey) if apikey is not None else ''})
                        elif config.credential_helper == 'huckle':
                            new_credentials[str(section_name)].append({str(name): str(value)})
                        else:
                            error = f"unknown credential helper configuration: {config.credential_helper}"
                            self._credentials = None
                            raise Exception(error)

                self._credentials = new_credentials
                return

        except Exception as e:
            error = f"huckle: unable to load credentials: {str(e)}"
            self._credentials = None
            raise Exception(error)

    def hcoak_find(self):
        try:
            self.parse_credentials()

            if not self._credentials:
                return None

            for section, cred_list in self._credentials.items():
                if section == config.auth_apikey_profile:
                    section_keyid = None
                    section_apikey = None
                    for cred in cred_list:
                        if 'keyid' in cred:
                            section_keyid = cred['keyid']
                        if 'apikey' in cred:
                            section_apikey = cred['apikey']

                    if section_keyid is not None and section_apikey is not None and not section_apikey == '':
                        return section_keyid, section_apikey

                    if section_apikey == '':
                        error = f"the apikey is empty for keyid {section_keyid}. validate expected keyring stored values retrieved for {config.url} and {section_keyid}."
                        raise Exception(error)

            error = f"unable to find keyid or apikey credentials for the [{config.auth_apikey_profile}] profile under {self.credentials_file_path}."
            raise Exception(error)

        except Exception as e:
            error = f"huckle: error retrieving credentials: {str(e)}"
            raise Exception(error)

    def find(self):
        try:
            self.parse_credentials()

            if not self._credentials:
                return None

            for section, cred_list in self._credentials.items():
                if section == config.auth_user_profile:
                    section_username = None
                    section_password = None
                    for cred in cred_list:
                        if 'username' in cred:
                            section_username = cred['username']
                        if 'password' in cred:
                            section_password = cred['password']

                    if section_username is not None and section_password is not None and not section_password == '':
                        return section_username, section_password

                    if section_password == '':
                        error = f"the password is empty for username {section_username}. validate expected keyring stored values retrieved for {config.url} and {section_username}."
                        raise Exception(error)

            error = f"unable to find username and password credentials for the [{config.auth_user_profile}] profile under {self.credentials_file_path}."
            raise Exception(error)

        except Exception as e:
            error = f"huckle: error retrieving credentials: {str(e)}"
            raise Exception(error)

    def update_credential(self, key, secret):
        try:
            config.create_file(self.credentials_file_path)
            with config.write_lock(self.credentials_file_path):
                log.debug("updating credentials file.")
                parser = ConfigParser(interpolation=None)
                log.debug(self.credentials_file_path)
                parser.read(self.credentials_file_path)

                is_apikey = secret.startswith('hcoak_')
                id_type = "keyid" if is_apikey else "username"
                value_type = "apikey" if is_apikey else "password"

                # default section created if the credentials file is empty first
                if not parser.sections():
                    # Create default section without number
                    new_section = f"default"
                    parser.add_section(new_section)
                    parser.set(new_section, id_type, key)
                    if config.credential_helper == 'keyring':
                        keyring.set_password(config.url, key, secret)
                    elif config.credential_helper == 'huckle':
                        parser.set(new_section, value_type, secret)
                else:
                    # Find existing credential and track highest section number
                    found_section = None
                    highest_num = 0
                    base_section = f"{id_type}_profile"

                    for section in parser.sections():
                        # Check if this is the credential we're updating
                        if parser.has_option(section, id_type) and parser.get(section, id_type) == key:
                            found_section = section

                        # Track highest section number
                        if section.startswith(base_section):
                            try:
                                section_num = int(section[len(base_section):] or 0)
                                highest_num = max(highest_num, section_num)
                            except ValueError:
                                continue

                    if found_section:
                        # Update existing credential
                        if config.credential_helper == 'keyring':
                            keyring.set_password(config.url, key, secret)
                        elif config.credential_helper == 'huckle':
                            parser.set(found_section, value_type, secret)
                    else:
                        # Create new section with next number
                        new_section = f"{base_section}{highest_num + 1}"
                        parser.add_section(new_section)
                        parser.set(new_section, id_type, key)

                        if config.credential_helper == 'keyring':
                            keyring.set_password(config.url, key, secret)
                        elif config.credential_helper == 'huckle':
                            parser.set(new_section, value_type, secret)

                with open(self.credentials_file_path, 'w') as configfile:
                    parser.write(configfile)
        except Exception as e:
            error = f"huckle: unable to set credentials: {str(e)}"
            raise Exception(error)

    @property
    def is_loaded(self):
        return self._credentials is not None



__version__ = "5.9.0"
dependencies = ["restnavigator==1.0.1",
                "requests[socks]==2.32.5",
                "configparser==5.3.0",
                "certifi",
                "portalocker==2.10.1",
                "keyring==25.7.0"]



from requests.auth import AuthBase
import base64


class HCOAKBearerAuth(AuthBase):
    def __init__(self, keyid, apikey):
        self.keyid = keyid
        self.apikey = apikey
        self.auth_string = f"{keyid}:{apikey}"

    def __call__(self, r):
        encoded = base64.b64encode(self.auth_string.encode()).decode()
        r.headers["authorization"] = f"Bearer {encoded}"
        return r
